//-------------------------------------------------------------------------------------------------
/// \file	BasicHwShader.cpp
/// \date	6-11-04
/// \author	Rob Bateman
/// \brief	This final version adds in the ray traced reflections into the shader
///
//-------------------------------------------------------------------------------------------------

#include "BasicSwShader.h"
#include <math.h>
#include <maya/MFnLightDataAttribute.h>
#include <maya/MRenderUtil.h>

#define MAKE_INPUT(attr)	\
	attr.setKeyable(true);	\
	attr.setStorable(true);	\
	attr.setReadable(true);	\
	attr.setWritable(true);

#define MAKE_OUTPUT(attr)		\
	attr.setKeyable(false); 	\
	attr.setStorable(false);	\
	attr.setReadable(true); 	\
	attr.setWritable(false);

const MTypeId BasicSwShader::m_TypeId( 0x70033 );
const MString BasicSwShader::m_TypeName( "basicSwShader" );


/// The diffuse colour of the material
MObject  BasicSwShader::m_aDiffuseColor;

/// The ambient colour of the material
MObject  BasicSwShader::m_aAmbientColor;
/// The ambient colour of the material
MObject  BasicSwShader::m_aSpecularColor;
/// The ambient colour of the material
MObject  BasicSwShader::m_aSpecularity;

/// The diffuse colour of the material
MObject  BasicSwShader::m_aDiffuseReflectivity;

/// The output colour of the material
MObject  BasicSwShader::m_aCosPower;

/// The output colour of the material
MObject  BasicSwShader::m_aOutColor;

/// the camera position
MObject  BasicSwShader::m_aPointCamera;

/// the surface normal
MObject  BasicSwShader::m_aNormalCamera;

/// the ray origin
MObject  BasicSwShader::m_aRayOrigin;

/// the ray direction
MObject  BasicSwShader::m_aRayDirection;

/// the light direction
MObject  BasicSwShader::m_aLightDirection;

/// the lights intensity
MObject  BasicSwShader::m_aLightIntensity;

// three booleans used specifically for render layers. Maya will set these to true or
// false to tell us whether we should calculate the various components.
MObject  BasicSwShader::m_aLightAmbient;
MObject  BasicSwShader::m_aLightDiffuse;
MObject  BasicSwShader::m_aLightSpecular;
MObject  BasicSwShader::m_aLightShadowFraction;
MObject  BasicSwShader::m_aPreShadowIntensity;

// additional lighting info
MObject  BasicSwShader::m_aLightBlindData;
MObject  BasicSwShader::m_aLightData;


MObject  BasicSwShader::m_aObjectId;
MObject  BasicSwShader::m_aRaySampler;
MObject  BasicSwShader::m_aRayDepth;

MObject  BasicSwShader::m_aReflectGain;

MObject  BasicSwShader::m_aTriangleNormalCamera;
MObject  BasicSwShader::m_aTranslucenceCoeff;

//
void* BasicSwShader::creator()
{
    return new BasicSwShader();
}

//
MStatus BasicSwShader::initialize()
{
	// we will use a numeric attribute function set to add all of
	// the attributes to the node description.
    MFnNumericAttribute nAttr;
    MFnLightDataAttribute lAttr;

	// create specularity attr
    m_aSpecularity = nAttr.create( "specularity", "spc", MFnNumericData::kFloat);
    MAKE_INPUT(nAttr);
    nAttr.setMin(0.0f);
    nAttr.setMax(1.0f);
    nAttr.setDefault(0.5f);

	// create refletion gain attr
    m_aReflectGain = nAttr.create( "reflectionGain", "rg", MFnNumericData::kFloat);
    MAKE_INPUT(nAttr);
    nAttr.setMin(0.0f);
    nAttr.setMax(1.0f);
    nAttr.setDefault(0.5f);

	// sampler info
    m_aTriangleNormalCamera = nAttr.createPoint( "triangleNormalCamera", "tn" );
    MAKE_INPUT(nAttr);
	nAttr.setDefault(1.0f, 1.0f, 1.0f);
	nAttr.setHidden(true);

	// translucence
    m_aTranslucenceCoeff = nAttr.create("translucenceCoeff", "tc",
									  MFnNumericData::kFloat);
    MAKE_INPUT(nAttr);


	// objectId
	m_aObjectId = nAttr.create( "objectId", "oi", MFnNumericData::kLong, 0.0 );
    nAttr.setStorable(false);
    nAttr.setHidden(true);
    nAttr.setReadable(false);

	// raySampler
	m_aRaySampler = nAttr.create("raySampler", "rtr", MFnNumericData::kLong,0.0);
    nAttr.setStorable(false);
    nAttr.setHidden(true) ;
    nAttr.setReadable(false) ;

	// rayDepth
	m_aRayDepth = nAttr.create( "rayDepth", "rd", MFnNumericData::kShort, 0.0 );
    nAttr.setStorable(false);
    nAttr.setHidden(true);
    nAttr.setReadable(false);

    // create the diffuse colour attribute
	m_aDiffuseColor =  nAttr.createColor( "color", "c");
	MAKE_INPUT(nAttr);
	nAttr.setDefault(0.7f, 0.7f, 0.7f);

    // create the ambient colour attribute
	m_aAmbientColor =  nAttr.createColor( "ambientColor", "ac");
	MAKE_INPUT(nAttr);
	nAttr.setDefault(0.4f, 0.4f, 0.4f);

	m_aSpecularColor =  nAttr.createColor( "specularColor", "sc");
	MAKE_INPUT(nAttr);
	nAttr.setDefault(1.0f, 1.0f, 1.0f);

	// a diffuse reflectivity value
	m_aDiffuseReflectivity = nAttr.create("diffuseReflectivity", "drfl", MFnNumericData::kFloat);
	MAKE_INPUT(nAttr);
	nAttr.setDefault(0.7f);

	// the cosine power, shininess value
	m_aCosPower = nAttr.create("shininess", "sh", MFnNumericData::kFloat);
	MAKE_INPUT(nAttr);
	nAttr.setDefault(10.0f);

    // create the output colour attribute
	m_aOutColor =  nAttr.createColor( "outColor", "oc");
	MAKE_OUTPUT(nAttr);

	// the camera's position
    m_aPointCamera = nAttr.createPoint( "pointCamera", "pc" );
    MAKE_INPUT(nAttr);
    nAttr.setDefault(1.0f, 1.0f, 1.0f);
    nAttr.setHidden(true);

	// the input surface normal
    m_aNormalCamera = nAttr.createPoint( "normalCamera", "n" );
    MAKE_INPUT(nAttr);
    nAttr.setDefault(1.0f, 1.0f, 1.0f);
    nAttr.setHidden(true);

	// rayOrigin
	MObject RayX = nAttr.create( "rayOx", "rxo", MFnNumericData::kFloat, 0.0 );
	MObject RayY = nAttr.create( "rayOy", "ryo", MFnNumericData::kFloat, 0.0 );
	MObject RayZ = nAttr.create( "rayOz", "rzo", MFnNumericData::kFloat, 0.0 );
	m_aRayOrigin = nAttr.create( "rayOrigin", "rog", RayX, RayY, RayZ );
    nAttr.setStorable(false);
    nAttr.setHidden(true);
    nAttr.setReadable(false);

	// rayDirection
	RayX = nAttr.create( "rayDirectionX", "rdx", MFnNumericData::kFloat, 1.0 );
	RayY = nAttr.create( "rayDirectionY", "rdy", MFnNumericData::kFloat, 0.0 );
	RayZ = nAttr.create( "rayDirectionZ", "rdz", MFnNumericData::kFloat, 0.0 );
	m_aRayDirection = nAttr.create( "rayDirection", "rad", RayX, RayY, RayZ );
    nAttr.setStorable(false);
    nAttr.setHidden(true);
    nAttr.setReadable(false);

	//--------------------------------------------------------------------------------------
	//	The following attributes will be added as a compound array attribute, lightDataArray.
	//
	//	Maya will fill this array with info about each light in the scene. We can then use
	//	that array within compute to calculate the lights value. In this example we shall
	//	keep it simple and just calculate the diffuse lighting....
	//

		// the direction of a light
		m_aLightDirection = nAttr.createPoint( "lightDirection", "ld" );
		MAKE_INPUT(nAttr);
		nAttr.setDefault(1.0f, 1.0f, 1.0f);
		nAttr.setHidden(true);

		// a lights intensity
		m_aLightIntensity = nAttr.createColor( "lightIntensity", "li" );
		MAKE_INPUT(nAttr);
		CHECK_MSTATUS ( nAttr.setDefault(1.0f, 1.0f, 1.0f) );
		nAttr.setHidden(true);

		// a flag to indicate that the ambient colour should be calculated
		m_aLightAmbient = nAttr.create( "lightAmbient", "la", MFnNumericData::kBoolean);
		MAKE_INPUT(nAttr);
		nAttr.setHidden(true);
		nAttr.setDefault(true);

		// a flag to indicate that the diffuse colour should be calculated
		m_aLightDiffuse = nAttr.create( "lightDiffuse", "ldf", MFnNumericData::kBoolean);
		MAKE_INPUT(nAttr);
		nAttr.setHidden(true);
		nAttr.setDefault(true);

		// a flag to indicate that the specular colour should be calculated
		m_aLightSpecular = nAttr.create( "lightSpecular", "ls", MFnNumericData::kBoolean);
		MAKE_INPUT(nAttr);
		nAttr.setHidden(true);
		nAttr.setDefault(true);

		// other data for a camera
		m_aLightBlindData = nAttr.create("lightBlindData", "lbld", MFnNumericData::kLong);
		MAKE_INPUT(nAttr);
		nAttr.setHidden(true);

		// the lights shadow value
		m_aLightShadowFraction = nAttr.create("lightShadowFraction", "lsf", MFnNumericData::kFloat);
		MAKE_INPUT(nAttr);
		nAttr.setHidden(true);

		// intensity before the shadow
		m_aPreShadowIntensity = nAttr.create("preShadowIntensity", "psi", MFnNumericData::kFloat);
		MAKE_INPUT(nAttr);
		nAttr.setHidden(true);

	//--------------------------------------------------------------------------------------

	// create a compound light array attribute
    m_aLightData = lAttr.create( "lightDataArray", "ltd",
                               m_aLightDirection,
							   m_aLightIntensity,
							   m_aLightAmbient,
                               m_aLightDiffuse,
							   m_aLightSpecular,
							   m_aLightShadowFraction,
                               m_aPreShadowIntensity,
                               m_aLightBlindData);
    lAttr.setArray(true);
    lAttr.setStorable(false);
    lAttr.setHidden(true);
    lAttr.setDefault(0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, true, true, false, 0.0f, 1.0f, 0);

	// add the various attributes to the node
	addAttribute(m_aAmbientColor);
	addAttribute(m_aDiffuseColor);
	addAttribute(m_aSpecularColor);
	addAttribute(m_aCosPower);
	addAttribute(m_aDiffuseReflectivity);
	addAttribute(m_aPointCamera);
	addAttribute(m_aNormalCamera);
	addAttribute(m_aRayOrigin);
	addAttribute(m_aRayDirection);
	addAttribute(m_aLightData);
	addAttribute(m_aOutColor);

	addAttribute(m_aObjectId);
	addAttribute(m_aRaySampler);
	addAttribute(m_aRayDepth);
	addAttribute(m_aReflectGain);
	addAttribute(m_aTriangleNormalCamera);
	addAttribute(m_aTranslucenceCoeff);

	// set the attribute dependencies

	attributeAffects(m_aObjectId,  			m_aOutColor);
	attributeAffects(m_aRaySampler,			m_aOutColor);
	attributeAffects(m_aRayDepth,  			m_aOutColor);
	attributeAffects(m_aReflectGain,  		m_aOutColor);
	attributeAffects(m_aTriangleNormalCamera,m_aOutColor);
	attributeAffects(m_aTranslucenceCoeff,  m_aOutColor);

	attributeAffects(m_aAmbientColor,  		m_aOutColor);
	attributeAffects(m_aDiffuseColor,  		m_aOutColor);
	attributeAffects(m_aSpecularColor,  	m_aOutColor);
	attributeAffects(m_aCosPower,  			m_aOutColor);
	attributeAffects(m_aDiffuseReflectivity,m_aOutColor);
	attributeAffects(m_aPointCamera,  		m_aOutColor);
	attributeAffects(m_aNormalCamera,  		m_aOutColor);
	attributeAffects(m_aRayOrigin,  		m_aOutColor);
	attributeAffects(m_aRayDirection,  		m_aOutColor);

	attributeAffects(m_aLightDirection,  	m_aOutColor);
	attributeAffects(m_aLightIntensity,  	m_aOutColor);
	attributeAffects(m_aLightAmbient,  		m_aOutColor);
	attributeAffects(m_aLightDiffuse,  		m_aOutColor);
	attributeAffects(m_aLightSpecular,  	m_aOutColor);
	attributeAffects(m_aLightShadowFraction,m_aOutColor);
	attributeAffects(m_aPreShadowIntensity, m_aOutColor);
	attributeAffects(m_aLightBlindData,  	m_aOutColor);

	// done
    return MS::kSuccess;
}


// computes the output colour. In a software shader this would be used to calculate an output
// colour for a pixel during rendering. There are 2 occasions when this takes place...
//
// 1. When the scene is rendered
// 2. When the sample swatch in the Attribute Editor gets rendered.
//
// In our case then, it will only get called when the swatch gets evaluated. Only software shaders
// can get rendered by the maya renderer...
//
MStatus BasicSwShader::compute (const MPlug& plug, MDataBlock& data)
{
	// Check that the requested recompute is one of the output values
	//
	if( (plug != m_aOutColor) &&
		(plug.parent() != m_aOutColor) )
		return MS::kUnknownParameter;

	// This will store the calculated diffuse lighting term
	double ambientTerm[3] = {0,0,0};

	// This will store the calculated diffuse lighting term
	double diffuseTerm[3] = {0,0,0};

	// This will store the calculated diffuse lighting term
	double specularTerm[3] = {0,0,0};

    // get sample surface shading parameters
    MFloatVector& surfaceNormal = data.inputValue( m_aNormalCamera ).asFloatVector();
    MFloatVector& cameraPosition = data.inputValue( m_aPointCamera ).asFloatVector();

	// we will use the diffuse reflectivity as a way of globally controlling the intensity
	// of the diffuse colour. This is use below in the diffuse calculation
	float diffuseReflectivity = data.inputValue(m_aDiffuseColor).asFloat();
	float cosPower = data.inputValue(m_aCosPower).asFloat();
	if( cosPower < 0 )
		cosPower = -cosPower;

	// used for raytracing api enhancement below
	MFloatVector normal = surfaceNormal;
	MFloatVector point = cameraPosition;

    // get the array of lights
    MArrayDataHandle lightData = data.inputArrayValue( m_aLightData );
    int numLights = lightData.elementCount();

 	// iterate through light list and accumulate the diffuse lighting values
	for( int count=1; count <= numLights; count++ )
	{
		// get a handle to the current light
		MDataHandle currentLight = lightData.inputValue();

		// get references to the light intensity and direction
		MFloatVector& lightIntensity = currentLight.child(m_aLightIntensity).asFloatVector();
		MFloatVector& lightDirection = currentLight.child(m_aLightDirection).asFloatVector();

		// calculate ambient component
		if( currentLight.child(m_aLightDiffuse).asBool() )
		{
			ambientTerm[0] += lightIntensity[0];
			ambientTerm[1] += lightIntensity[1];
			ambientTerm[2] += lightIntensity[2];
		}

		// Find the blind data for the light
		int& blindData = currentLight.child( m_aLightBlindData ).asInt();
		if ( blindData == 0 )
		{
			// if diffuse is to be calculated, add in diffuse term
			if( currentLight.child(m_aLightDiffuse).asBool() )
			{
				// dot product the light direction and surface normal, L.N
				float LN = lightDirection * surfaceNormal;

				// calculate only if facing light
				if( LN > 0.0f )
				{
					diffuseTerm[0] += lightIntensity[0] * LN * diffuseReflectivity;
					diffuseTerm[1] += lightIntensity[1] * LN * diffuseReflectivity;
					diffuseTerm[2] += lightIntensity[2] * LN * diffuseReflectivity;
			    }

				//
				cameraPosition.normalize();

				// calculate only if facing light
				if( LN > 0.0f )
				{
				    float RV = ( ( (2*surfaceNormal) * LN ) - lightDirection ) * cameraPosition;

					if( RV > 0.0 )
						RV = 0.0;
					if( RV < 0.0 )
						RV = -RV;

					float s = powf( RV, cosPower );

					specularTerm[0] += lightIntensity[0] * s;
					specularTerm[1] += lightIntensity[1] * s;
					specularTerm[2] += lightIntensity[2] * s;
				}
			}
        }
        else
        {
			// calculate LN using reflected rays
			float LN = MRenderUtil::diffuseReflectance( blindData, lightDirection, point, surfaceNormal, true );

			// calculate only if facing light
			if( LN > 0.0f )
			{
			     diffuseTerm[0] += lightIntensity[0] * LN * diffuseReflectivity;
			     diffuseTerm[1] += lightIntensity[1] * LN * diffuseReflectivity;
			     diffuseTerm[2] += lightIntensity[2] * LN * diffuseReflectivity;
			}

			cameraPosition.normalize();

			// if the specular is required...
			if ( currentLight.child(m_aLightSpecular).asBool() )
			{
				MFloatVector specLightDirection = lightDirection;
				MDataHandle directionH = data.inputValue( m_aRayDirection );
				MFloatVector direction = directionH.asFloatVector();
				float lightAttenuation = 1.0;

				specLightDirection = MRenderUtil::maximumSpecularReflection( blindData, lightDirection, point, surfaceNormal, direction );
				lightAttenuation = MRenderUtil::lightAttenuation( blindData, point, surfaceNormal, false );

				// Are we facing the light
				if ( specLightDirection * surfaceNormal > 0.0f )
				{
					MFloatVector rv = 2 * surfaceNormal * ( surfaceNormal * direction ) - direction;
					float s = powf( rv * specLightDirection, cosPower );

					specularTerm[0] += lightIntensity[0] * s * lightAttenuation;
					specularTerm[1] += lightIntensity[1] * s * lightAttenuation;
					specularTerm[2] += lightIntensity[2] * s * lightAttenuation;
				}
			 }
		}

		// next light please
		if( !lightData.next() )
			break;
	}

	// use the handle to get the diffuse color value
	const MFloatVector& ambientColor = data.inputValue(m_aAmbientColor).asFloatVector();
	const MFloatVector& diffuseColor = data.inputValue(m_aDiffuseColor).asFloatVector();
	const MFloatVector& specularColor = data.inputValue(m_aSpecularColor).asFloatVector();


	// again, get hold of a handle to the output attribute... This time
	// however we get an outputHandle.
	//
	MDataHandle outColorHandle = data.outputValue( m_aOutColor );

	// get a reference to the actual data
    MFloatVector & outColor = outColorHandle.asFloatVector();

	// set the output colour, diffuseTerm * diffuse colour
	outColor[0] = static_cast<float>(
				  ambientColor[0]  * ambientTerm[0] +
				  diffuseColor[0]  * diffuseTerm[0] +
				  specularColor[0] * specularTerm[0]);
	outColor[1] = static_cast<float>(
				  ambientColor[1]  * ambientTerm[1] +
				  diffuseColor[1]  * diffuseTerm[1] +
				  specularColor[1] * specularTerm[1]);
	outColor[2] = static_cast<float>(
				  ambientColor[2]  * ambientTerm[2] +
				  diffuseColor[2]  * diffuseTerm[2] +
				  specularColor[2] * specularTerm[2]);

	float reflectGain = data.inputValue( m_aReflectGain ).asFloat();

	// add the reflection color
	if (reflectGain > 0.0) {

		MStatus status;

		// required attributes for using raytracer
		// origin, direction, sampler, depth, and object id.
		//
		MFloatVector origin 	= data.inputValue( m_aRayOrigin ).asFloatVector();
		MFloatVector direction  = data.inputValue( m_aRayDirection ).asFloatVector();

		int samplerPtr = data.inputValue( m_aRaySampler ).asLong();

		short depth = data.inputValue( m_aRayDepth ).asShort();

		int objId = data.inputValue( m_aObjectId ).asLong();

		MFloatVector reflectColor;
		MFloatVector reflectTransparency;

		MFloatVector& triangleNormal = data.inputValue( m_aTriangleNormalCamera ).asFloatVector();

		// compute reflected ray
		MFloatVector l = -direction;
		float dot = l * normal;

		if( dot < 0.0 )
			dot = -dot;

		// reflection ray
		MFloatVector refVector = 2 * normal * dot - l;
		float dotRef = refVector * triangleNormal;

		if( dotRef < 0.0 ) {
		    const float s = 0.01f;
			MFloatVector mVec = refVector - dotRef * triangleNormal;
			mVec.normalize();
			refVector = mVec + s * triangleNormal;
		}
		refVector.normalize();

		MRenderUtil::raytrace(
				point,    		//  origin
				refVector,  	//  direction
				objId,			//  object id
				samplerPtr, 	//  sampler info
				depth,			//  ray depth
				reflectColor,	//	output color and transp
				reflectTransparency);

		// add in the reflection color
		outColor[0] += reflectGain * (reflectColor[0]);
		outColor[1] += reflectGain * (reflectColor[1]);
		outColor[2] += reflectGain * (reflectColor[2]);

	}

	// clean the plug, ie flag it as clean so that maya does not re-evaluate this attribute
	outColorHandle.setClean();

	return MS::kSuccess;
}
