//-------------------------------------------------------------------------------------------------
/// \file	BasicSwShader.h
/// \date	6-11-04
/// \author	Rob Bateman
/// \brief	This final version adds in the ray traced reflections into the shader
///
//-------------------------------------------------------------------------------------------------

#ifndef __BASIC_SW_SHADER__H__
#define __BASIC_SW_SHADER__H__

	#ifdef WIN32
		#define NT_PLUGIN
		#pragma once
		#define WIN32_LEAN_AND_MEAN
		#include <windows.h>
	#endif

	#include <maya/MIOStream.h>
	#include <maya/MString.h>
	#include <maya/MPlug.h>
	#include <maya/MDataBlock.h>
	#include <maya/MDataHandle.h>
	#include <maya/MFnNumericAttribute.h>
	#include <maya/MFloatVector.h>
	#include <maya/MPxNode.h>

	//---------------------------------------------------------------------------------------------
	/// \brief
	///
	class BasicSwShader
		: public MPxNode
	{
	public:

		/// \brief	The compute function is called to calculate the value of a specific
		/// 		output attribute. This usually only happens when an input value changes
		/// 		on which the output was dependant on; or when the node is forced to be
		/// 		recomputed because it has been flagged as dirty.
		/// \param	plug	-	the output attribute to calculate
		/// \param	data	-	the data from which you MUST read the attributes back from.
		///	\return	Did it succeed or fail?
		///
		/// \note	YOU MUST ONLY EVER READ THE NODE'S ATTRIBUTES BACK FROM THE DATA BLOCK
		/// 		GIVEN TO YOU. YOU MUST NOT ATTEMPT TO USE ANY FUNCTION SETS, AND YOU MUST
		/// 		NOT ACCESS ANY OTHER NODES WHILST IN THE COMPUTE FUNCTION.
		///
		/// 		THE COMPUTE FUNCTION IS THE ONLY FUNCTION ON OUR NODE, THAT IS LIKELY TO
		/// 		BE USED DURING RENDERING. BY ACCESSING ONLY THE ATTRIBUTES VIA THE DATA
		/// 		BLOCK, MAYA ENSURES THAT RENDERING REMAINS THREAD SAFE. DUPLICATE
		/// 		DATA BLOCKS CAN BE PASSED TO TWO OR MORE PROCESSORS. THIS MEANS THAT NO
		/// 		SHARED DATA WILL BE ACCESSED BY DIFFERENT THREADS. ie, EACH THREAD RECIEVES
		/// 		IT'S OWN COPY OF THE DATA.
		///
		/// 		IF YOU DO START ACCESSING GLOBAL DATA, OR DATA FROM OTHER NODES WITHIN THE
		/// 		COMPUTE FUNCTION - IT WILL PROBABLY WORK WHILST YOU'RE WORKING INSIDE MAYA,
		/// 		HOWEVER - IT IS LIABLE TO CRASH WHEN RENDERING, ie WHEN MULTIPLE PROCESSORS
		/// 		ARE USED.
		///
		virtual MStatus compute( const MPlug&, MDataBlock& );

	public:

		/// \brief	This function returns a new instance of our shader
		/// \return	A new shader instance
		///
		static  void *  creator();

		/// \brief	This function sets up the attributes on the shader node
		/// \return	ok or fail
		///
		static  MStatus initialize();

	public:

		// 	------------ The Type Info ------------

		/// The specific ID for this node type
		const static  MTypeId m_TypeId;

		/// The type name for this new node type
		const static  MString m_TypeName;

	private:

		// 	------------ The material properties ------------

		/// The ambient colour of the material
		static MObject m_aAmbientColor;

		/// The diffuse colour of the material
		static MObject m_aDiffuseColor;

		/// The specular colour of the material
		static MObject m_aSpecularColor;
   		static MObject m_aSpecularity;

		/// The specular colour of the material
		static MObject m_aCosPower;

		/// The diffuse colour of the material
		static MObject m_aDiffuseReflectivity;



		// 	------------ Surface Shading Properties ------------

		/// the camera position
		static MObject m_aPointCamera;

		/// the surface normal
		static MObject m_aNormalCamera;



		// 	------------ Light Values ------------

		/// the light direction
		static MObject m_aLightDirection;

		/// the lights intensity
		static MObject m_aLightIntensity;



		// three booleans used specifically for render layers. Maya will set these to true or
		// false to tell us whether we should calculate the various components.
		static MObject m_aLightAmbient;
		static MObject m_aLightDiffuse;
		static MObject m_aLightSpecular;
		static MObject m_aLightShadowFraction;
		static MObject m_aPreShadowIntensity;

		// additional lighting info
		static MObject m_aLightBlindData;
		static MObject m_aLightData;
		static MObject m_aRayOrigin;
		static MObject m_aRayDirection;
		static MObject m_aObjectId;
		static MObject m_aRaySampler;
		static MObject m_aRayDepth;
		static MObject m_aReflectGain;
		static MObject m_aTriangleNormalCamera;
	    static MObject m_aTranslucenceCoeff;

		// 	------------ Output ------------

		/// The output colour for the material
		static MObject m_aOutColor;
	};


#endif
